#include <vector>

using namespace std;

#include "I2C.h"
#include <stdbool.h>
#include <stdio.h>
#include <fcntl.h>
#include <io.h>
#include <stdarg.h>
#include <windows.h>
#include <time.h>
#include <errno.h>


//#define WITH_DEBUG

/*********************************************************
		Globals signal setup
*********************************************************/


//static FILE *logfile = NULL;


const char *modeinfo_names[MODEINFO_MAX] = {
	"MAX_BUS",
	"MAX_GROUP",
	"MAX_MODE",
	"3",
	"GETNAME"
};


const int bustable[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

//Define groups (4 signals)
const struct groupinfo groupinfo[] = { { "I2C", 0, 0, 0, 0, 2, 0 },
                                 { "Traffic" , 0, 0, 2, 1, 0x80, sizeof("Traffic")-1 } };

const struct businfo businfo[] = { { 0, 0, 0, 0, 0, 0x20000, NULL, 0 } };

//Define settings
//const char *onoff[] = { "0", "1", NULL, NULL };
//const char *sel_mode[] = { "LOW", "HIGH", "Don't care", NULL };
//const char *bit_widths[]={"4", "5", "6", "7", "8", "9","10","11","12","13","14","15","16", NULL};

//const struct modeinfo modeinfo[] = { { "CPOL", onoff, 0, 1 },
																//{ "CPHA", onoff, 0, 1 },
																//{ "SEL", sel_mode, 0, 2 },
																//{ "BITS", bit_widths, 0, 12 }};

/*********************************************************
		SPI analysis data
*********************************************************/

//aray to translate between settings and bit width
const int bit_width_idx[]={4,5,6,7,8,9,10,11,12,13,14,15,16};

//static TI2CData I2CData[1];		//found data

//settings
//static int set_cpol;
//static int set_cpha;
//static int set_sel;
//static int set_bits;		//how many bits are in one sequence

static int processing_done;


TVSeqData SeqDataVector;		//vector with results

/*********************************************************
		Helpers
*********************************************************/
#ifdef WITH_DEBUG
static void LogDebug(struct pctx *pctx, int level, const char *fmt, ...)
{
        char buf[4096];
	va_list ap;
        va_start(ap, fmt);
        vsnprintf(buf, sizeof(buf), fmt, ap);
        va_end(ap);
        OutputDebugString(buf);
}

#else

#define LogDebug(...)

#endif


/*********************************************************
		DLL functions
*********************************************************/


struct pctx  *ParseReinit(struct pctx *pctx, struct lactx *lactx, struct lafunc *func)
{
	//Called upon DLL startup and refreshing the data
	struct pctx *ret;

	SeqDataVector.clear();
	processing_done=0;

	//Check if already initialized
	if (pctx!=NULL)
		{
		return pctx;	//already initialized -> exit
		}

	if (!(ret=(struct pctx*) func->rda_calloc(1, sizeof(struct pctx))))
		{
		func->LAError(0, 9, "Out of Memory");
		return NULL;
		}

	ret->lactx=lactx;
	ret->func.LAGroupValue=func->LAGroupValue;
	ret->func.rda_calloc=func->rda_calloc;
	ret->func.rda_free=func->rda_free;
	ret->func.LAInfo=func->LAInfo;

	//defaults
//	set_cpol=0;
//	set_cpha=0;
//	set_sel=2;
//	set_bits=8;
	SeqDataVector.clear();
	processing_done=0;

	LogDebug(ret, 9, "Initialization finished");
	return ret;
}

int  ParseExtInfo_(struct pctx *pctx)
{
	LogDebug(pctx, 9, "%s", __FUNCTION__);
	return 0;
}

int  ParseFinish(struct pctx *pctx)
{
	LogDebug(pctx, 9, "%s", __FUNCTION__);
	SeqDataVector.clear();
	pctx->func.rda_free(pctx);
	return 0;
}

struct sequence *ParseSeq(struct pctx *pctx, int initseq)
{
	//This function is called for each data sample. The sample
	//is identified by initseq. The order is more or less random.
	//When this function is called for the first time, all data
	//is procesed and stored in a vector. After that the
	//information is read from the vector.

	int last_value=I2C_MASK;	//previous sample


	TSeqData	SeqData;	//array to hold the new vector element.
	TVSeqData::iterator it;

	struct sequence *seqinfo=NULL;
	int value, firstseq, lastseq, seq, bitcnt, is_address;

	int sda,scl;
	int last_sda,last_scl;
	uint32_t i2c_data;

	EI2CState I2CState=I2C_IDLE;
	bitcnt=0;
	is_address=1;
	i2c_data=0;


	if (pctx == NULL)
		{
		LogDebug(pctx, 9, "pctx NULL");
		return NULL;
		}

	if (processing_done==0)
		{
		processing_done=1;

		//process the input data
		firstseq=pctx->func.LAInfo(pctx->lactx, TLA_INFO_FIRST_SEQUENCE, -1);
		lastseq=pctx->func.LAInfo(pctx->lactx, TLA_INFO_LAST_SEQUENCE, -1);

		LogDebug(pctx, 9, "initseq: %d, firstseq: %d, last seq: %d", initseq,
			firstseq, lastseq);

		//now loop through all the samples
		for (seq=firstseq; seq <= lastseq; seq++)
			{
			value=pctx->func.LAGroupValue(pctx->lactx, seq, 0);

			if (value!=last_value)
      	{
				//signals have changed
				scl=value & I2C_SCL;
				sda=value & I2C_SDA;

				LogDebug(pctx, 0, "%s: sequence %d sda: %d scl: %d", __FUNCTION__, seq, sda, scl);

				last_scl=last_value & I2C_SCL;
				last_sda=last_value & I2C_SDA;

				switch(I2CState)
					{
					case I2C_IDLE:
						if ((last_sda!=0) && (sda==0) && (scl!=0))
							{
							//sda goes low while scl=1 -> start condition
							memset(&SeqData,0, sizeof(SeqData));
							snprintf(SeqData.seq_data.text, sizeof(SeqData.seq_data.text), "I2C Start");
							SeqData.seq_data.flags=2;
							LogDebug(pctx, 0, "%s: sequence %s", __FUNCTION__, SeqData.seq_data.text);
							SeqData.seq_number=seq;
							SeqDataVector.push_back(SeqData);	//add data to vector

							bitcnt=0;
							is_address=1;
							i2c_data=0;
							I2CState=I2C_DATA;
							}
						break;

					case I2C_DATA:
						if ((last_sda==0) && (sda!=0) && (scl!=0))
							{
							//stop condition
							memset(&SeqData,0, sizeof(SeqData));
							snprintf(SeqData.seq_data.text, sizeof(SeqData.seq_data.text), "I2C Stop");
							SeqData.seq_data.flags=4;
							LogDebug(pctx, 0, "%s: sequence %s", __FUNCTION__, SeqData.seq_data.text);
							SeqData.seq_number=seq;
							SeqDataVector.push_back(SeqData);	//add data to vector
							I2CState=I2C_IDLE;
							}

						if ((last_sda!=0) && (sda==0) && (scl!=0))
							{
							//sda goes low while scl=1 -> restart condition
							memset(&SeqData,0, sizeof(SeqData));
							snprintf(SeqData.seq_data.text, sizeof(SeqData.seq_data.text), "I2C Start");
							SeqData.seq_data.flags=2;
							LogDebug(pctx, 0, "%s: sequence %s", __FUNCTION__, SeqData.seq_data.text);
							SeqData.seq_number=seq;
							SeqDataVector.push_back(SeqData);	//add data to vector

							is_address=1;
							bitcnt=0;
							i2c_data=0;
							}


						if ((last_scl==0) && (scl!=0))
							{
							//rising edge on scl -> data bit

							LogDebug(pctx, 0, "%s: bit %d", __FUNCTION__, bitcnt);

							i2c_data<<=1;	//shift
							if (sda!=0) i2c_data|=1;	//shift a 1 in

							bitcnt++;
							if (bitcnt>=8)
								{
								I2CState=I2C_ACK;
								}
							}
						break;

					case I2C_ACK:
						if ((last_sda==0) && (sda!=0) && (scl!=0))
							{
							//stop condition
							memset(&SeqData,0, sizeof(SeqData));
							snprintf(SeqData.seq_data.text, sizeof(SeqData.seq_data.text), "I2C Stop");
							SeqData.seq_data.flags=4;
							LogDebug(pctx, 0, "%s: sequence %s", __FUNCTION__, SeqData.seq_data.text);
							SeqData.seq_number=seq;
							SeqDataVector.push_back(SeqData);	//add data to vector
							I2CState=I2C_IDLE;
							}

						if ((last_scl==0) && (scl!=0))
							{
							LogDebug(pctx, 0, "%s: ack %d", __FUNCTION__, bitcnt);
							//rising edge on scl -> ack bit
							if (is_address!=0)
								{
								//we are in the address phase
								if ((i2c_data == 0x92) || (i2c_data == 0x93)) {
										snprintf(SeqData.seq_data.text, sizeof(SeqData.seq_data.text), "%s 0x%02x %s",
										i2c_data & 0x01 ? "MCU READ" : "MCU WRITE",
										i2c_data & 0xfe,
										(sda!=0) ? "NAK" : "ACK");
								} else {
										snprintf(SeqData.seq_data.text, sizeof(SeqData.seq_data.text), "%s 0x%02x %s",
										i2c_data & 0x01 ? "READ" : "WRITE",
										i2c_data & 0xfe,
										(sda!=0) ? "NAK" : "ACK");
									}
								}
							else
								{
								snprintf(SeqData.seq_data.text, sizeof(SeqData.seq_data.text), "DATA 0x%02x %s", i2c_data, (sda!=0) ? "NAK" : "ACK");
								}
							SeqData.seq_data.flags=8;
							LogDebug(pctx, 0, "%s: sequence %s", __FUNCTION__, SeqData.seq_data.text);
							SeqData.seq_number=seq;
							SeqDataVector.push_back(SeqData);	//add data to vector
							is_address=0;

							//reset bitcount and shift reg.
							i2c_data=0;
							bitcnt=0;

							I2CState=I2C_DATA;
							}
						break;

					default:
						I2CState=I2C_IDLE;
						break;

					}
				last_value=value;	//save previous value to detect edges
      	}
			}

		//sort out the text pointers
		it=SeqDataVector.begin();
		while(it!=SeqDataVector.end())
			{
			it->seq_data.textp=it->seq_data.text;
			it->seq_data.text2=it->seq_data.text2_buf;
			it++;
			}
		}


	//This could be done smarter if we keep track of the order the LA is
	//requesting the sequences. The next item we need is either the next
	//or the previous.
	it=SeqDataVector.begin();
	while( (it!=SeqDataVector.end()) && (seqinfo==NULL))
		{
		if (it->seq_number==initseq)
			{
			seqinfo=&(it->seq_data);

			LogDebug(pctx, 9, "%s: seq: %d text: [%s]", __FUNCTION__,
					initseq, seqinfo->textp);
			}
		it++;
		}

	return seqinfo;
}


int  ParseMarkNext(struct pctx *pctx, int seq, int a3)
{
	LogDebug(pctx, 9, "%s: sequence %d, a3 %d", __FUNCTION__, seq, a3);
	return 0;
}


int  ParseMarkSet(struct pctx *pctx, int seq, int a3)
{
	LogDebug(pctx, 9, "%s", __FUNCTION__);
	return 0;
}


int  ParseMarkGet(struct pctx *pctx, int seq)
{
	LogDebug(pctx, 9, "%s: sequence %d", __FUNCTION__, seq);
	return seq;
}


int  ParseMarkMenu(struct pctx *pctx, int seq, int a3, int a4, int a5)
{
	LogDebug(pctx, 0, "%s: sequence %d, a3: %08x, a4: %08x, a5: %08x", __FUNCTION__, seq, a3, a4, a5);
	return 0;
}


int  ParseInfo(struct pctx *pctx, unsigned int request)
{
	LogDebug(pctx, 6, "%s: %s", __FUNCTION__,
		request > ARRAY_SIZE(modeinfo_names) ? "invalid" : modeinfo_names[request]);

	switch(request) {
		case MODEINFO_MAX_BUS:
			return ARRAY_SIZE(businfo);
		case MODEINFO_MAX_GROUP:
			return ARRAY_SIZE(groupinfo);
		case MODEINFO_GETNAME:
			return (int)"I2C";
		case 3:
			return 1;
		case MODEINFO_MAX_MODE:
			return 0;	//no modeinfo
			//return ARRAY_SIZE(modeinfo);
		default:
			LogDebug(pctx, 9, "%s: invalid request: %d", __FUNCTION__, request);
			return 0;
	}	
	return 0;
}


struct businfo  *ParseBusInfo(struct pctx *pctx, uint16_t bus)
{
	LogDebug(pctx, 6, "%s: %08x", __FUNCTION__, bus);

	//This function gets called before parsing sequences after a
	//new acquisition so lets use it as a reset for parsing data
	SeqDataVector.clear();
	processing_done=0;

	if (bus >= ARRAY_SIZE(businfo))
		return NULL;
	return (struct businfo*) businfo+bus;
}


struct groupinfo  *ParseGroupInfo(struct pctx *pctx, uint16_t group)
{
	LogDebug(pctx, 9, "%s: %08x", __FUNCTION__, group);

	if (group > ARRAY_SIZE(groupinfo))
			return NULL;
	return (struct groupinfo*) groupinfo+group;
}


struct modeinfo  *ParseModeInfo(struct pctx *pctx, uint16_t mode)
{
	LogDebug(pctx, 9, "%s: %d", __FUNCTION__, mode);

	return NULL;	//no modeinfo
	/*
	if (mode > ARRAY_SIZE(modeinfo))
			return NULL;
	return (struct modeinfo*) modeinfo+mode;
	*/
}


int  ParseModeGetPut(struct pctx *pctx, int mode, int value, int request)
{
	//int i;
	LogDebug(pctx, 9, "%s: mode: %d value: %d request: %d ", __FUNCTION__,
			mode, value, request);
/*
	//mode is an index to the setting
	//value is the selected index
	//request= 0=read, 1=write, 2=write/read (in that order!)

	//write values?
	if ((request == 1) || (request==2))
		{
		//set values
		switch (mode)
			{
			case 0:
				//CPOL setting
				set_cpol=value;
				break;

			case 1:
				//CPHA setting
				set_cpha=value;
				break;

			case 2:
				//SEL setting
				set_sel=value;
				break;

			case 3:
				//bit width setting
				if ((value>=0) && (value<ARRAY_SIZE(bit_width_idx)))	//check range
					{
					set_bits=bit_width_idx[value];
					}
				break;

			default:
				break;
			}
		}

	//read values?
	if ((request == 0) || (request==2))
		{
		//request=0
		//read values
		value=0;	//choose default value
		switch (mode)
			{
			case 0:
				//CPOL setting
				value=set_cpol;
				break;

			case 1:
				//CPHA setting
				value=set_cpha;
				break;

			case 2:
				//SEL setting
				value=set_sel;
				break;

			case 3:
				//bit width setting
				//This one is tricky; we must find the index
				for(i=0; i<ARRAY_SIZE(bit_width_idx); i++ )
					{
					if (bit_width_idx[i]==set_bits)
						{
						value=i;
						break;
						}
					}
				break;

			default:
				value=0;
				break;
			}
		}
	LogDebug(pctx, 9, "%s: cpol: %d cpha: %d sel: %d bits: %d ", __FUNCTION__,
			set_cpol,  set_cpha, set_sel, set_bits);

*/
	return 0;
	//return value;
}


int  ParseDisasmReinit(struct pctx *pctx, int request)
{
	//Initialize parsing??
	LogDebug(pctx, 9, "%s: pctx %08x, request=%d", __FUNCTION__, pctx, request);

	return 1;
}
